"""alter_train_profile_datasetid_nullable

Revision ID: ed56a42729b7
Revises: c1ee06d84c12
Create Date: 2021-07-26 17:21:01.231388

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ed56a42729b7'
down_revision = 'c1ee06d84c12'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('local_train_profile', schema=None) as batch_op:
        batch_op.alter_column('dataset_id',
               existing_type=sa.VARCHAR(length=32),
               nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('local_train_profile', schema=None) as batch_op:
        batch_op.alter_column('dataset_id',
               existing_type=sa.VARCHAR(length=32),
               nullable=False)

    # ### end Alembic commands ###
