"""add train hpo id hpo final metric

Revision ID: 6dfd859f129e
Revises: 848203d98d73
Create Date: 2021-05-18 10:54:32.073962

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6dfd859f129e'
down_revision = '848203d98d73'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('local_train_task', schema=None) as batch_op:
        batch_op.add_column(sa.Column('hpo_final_metric', sa.DECIMAL(), nullable=True))
        batch_op.add_column(sa.Column('hpo_id', sa.String(length=8), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('local_train_task', schema=None) as batch_op:
        batch_op.drop_column('hpo_id')
        batch_op.drop_column('hpo_final_metric')

    # ### end Alembic commands ###
